@@ -81,7 +81,8 @@
 #include <ConfigKit/SubComponentUtils.h>
 #include <ServerKit/Server.h>
 #include <ServerKit/AcceptLoadBalancer.h>
-#include <MessageReadersWriters.h>
+#include <AppTypeDetector/Detector.h>
+#include <IOTools/MessageSerialization.h>
 #include <FileDescriptor.h>
 #include <ResourceLocator.h>
 #include <BackgroundEventLoop.cpp>
@@ -90,7 +91,7 @@
 #include <Exceptions.h>
 #include <Utils.h>
 #include <Utils/Timer.h>
-#include <Utils/MessageIO.h>
+#include <IOTools/MessageIO.h>
 #include <Core/OptionParser.h>
 #include <Core/Controller.h>
 #include <Core/ApiServer.h>
@@ -98,6 +99,7 @@
 #include <Core/ConfigChange.h>
 #include <Core/ApplicationPool/Pool.h>
 #include <Core/SecurityUpdateChecker.h>
+#include <Core/TelemetryCollector.h>
 #include <Core/AdminPanelConnector.h>
 
 using namespace boost;
@@ -165,6 +167,7 @@ namespace Core {
 		oxt::thread *prestarterThread;
 
 		SecurityUpdateChecker *securityUpdateChecker;
+		TelemetryCollector *telemetryCollector;
 		AdminPanelConnector *adminPanelConnector;
 		oxt::thread *adminPanelConnectorThread;
 
@@ -175,6 +178,7 @@ namespace Core {
 			  shutdownCounter(0),
 			  prestarterThread(NULL),
 			  securityUpdateChecker(NULL),
+			  telemetryCollector(NULL),
 			  adminPanelConnector(NULL),
 			  adminPanelConnectorThread(NULL)
 			  /*******************/
@@ -190,6 +194,7 @@ namespace Core {
 			delete adminPanelConnectorThread;
 			delete adminPanelConnector;
 			delete securityUpdateChecker;
+			delete telemetryCollector;
 
 			/*******************/
 			/*******************/
@@ -211,6 +216,7 @@ namespace Core {
 
 using namespace Passenger::Core;
 
+static WrapperRegistry::Registry *coreWrapperRegistry;
 static Schema *coreSchema;
 static ConfigKit::Store *coreConfig;
 static WorkingObjects *workingObjects;
@@ -252,48 +258,83 @@ initializeSingleAppMode() {
 	}
 
 	WorkingObjects *wo = workingObjects;
-	string appType, startupFile;
+	string appType, startupFile, appStartCommand;
 	string appRoot = coreConfig->get("single_app_mode_app_root").asString();
 
-	if (coreConfig->get("single_app_mode_app_type").isNull()) {
-		P_DEBUG("Autodetecting application type...");
-		AppTypeDetector detector(NULL, 0);
-		PassengerAppType appTypeEnum = detector.checkAppRoot(appRoot);
-		if (appTypeEnum == PAT_NONE || appTypeEnum == PAT_ERROR) {
-			fprintf(stderr, "ERROR: unable to autodetect what kind of application "
-				"lives in %s. Please specify information about the app using "
-				"--app-type and --startup-file, or specify a correct location to "
-				"the application you want to serve.\n"
-				"Type '" SHORT_PROGRAM_NAME " core --help' for more information.\n",
-				appRoot.c_str());
-			exit(1);
-		}
-
-		appType = getAppTypeName(appTypeEnum);
-	} else {
-		appType = coreConfig->get("single_app_mode_app_type").asString();
-	}
-
-	if (coreConfig->get("single_app_mode_startup_file").isNull()) {
-		startupFile = appRoot + "/" + getAppTypeStartupFile(getAppType(appType));
-	} else {
-		startupFile = coreConfig->get("single_app_mode_startup_file").asString();
-	}
-	if (!fileExists(startupFile)) {
-		fprintf(stderr, "ERROR: unable to find expected startup file %s."
-			" Please specify its correct path with --startup-file.\n",
-			startupFile.c_str());
+	if (!coreConfig->get("single_app_mode_app_type").isNull()
+	 && !coreConfig->get("single_app_mode_app_start_command").isNull())
+	{
+		fprintf(stderr, "ERROR: it is not allowed for both --app-type and"
+			" --app-start-command to be set.\n");
 		exit(1);
 	}
 
+	if (!coreConfig->get("single_app_mode_app_start_command").isNull()) {
+		// The config specified that this is a generic app or a Kuria app.
+		appStartCommand = coreConfig->get("single_app_mode_app_start_command").asString();
+	} else if (coreConfig->get("single_app_mode_app_type").isNull()) {
+		// Autodetect whether this is generic app, Kuria app or auto-supported app.
+		P_DEBUG("Autodetecting application type...");
+		AppTypeDetector::Detector detector(*coreWrapperRegistry);
+		AppTypeDetector::Detector::Result detectorResult = detector.checkAppRoot(appRoot);
+
+		if (!detectorResult.appStartCommand.empty()) {
+			// This is a generic or Kuria app.
+			appStartCommand = detectorResult.appStartCommand;
+		} else {
+			// This is an auto-supported app.
+			if (coreConfig->get("single_app_mode_app_type").isNull()) {
+				if (detectorResult.isNull()) {
+					fprintf(stderr, "ERROR: unable to autodetect what kind of application "
+						"lives in %s. Please specify information about the app using "
+						"--app-type, --startup-file and --app-start-command, or specify a "
+						"correct location to the application you want to serve.\n"
+						"Type '" SHORT_PROGRAM_NAME " core --help' for more information.\n",
+						appRoot.c_str());
+					exit(1);
+				}
+				appType = detectorResult.wrapperRegistryEntry->language;
+			} else {
+				appType = coreConfig->get("single_app_mode_app_type").asString();
+			}
+		}
+	} else {
+		// This is an auto-supported app.
+		appType = coreConfig->get("single_app_mode_app_type").asString();
+	}
+
+	if (!appType.empty()) {
+		if (coreConfig->get("single_app_mode_startup_file").isNull()) {
+			const WrapperRegistry::Entry &entry = coreWrapperRegistry->lookup(appType);
+			if (entry.defaultStartupFiles.empty()) {
+				startupFile = appRoot + "/";
+			} else {
+				startupFile = appRoot + "/" + entry.defaultStartupFiles[0];
+			}
+		} else {
+			startupFile = coreConfig->get("single_app_mode_startup_file").asString();
+		}
+		if (!fileExists(startupFile)) {
+			fprintf(stderr, "ERROR: unable to find expected startup file %s."
+				" Please specify its correct path with --startup-file.\n",
+				startupFile.c_str());
+			exit(1);
+		}
+	}
+
 	wo->singleAppModeConfig["app_root"] = appRoot;
-	wo->singleAppModeConfig["app_type"] = appType;
-	wo->singleAppModeConfig["startup_file"] = startupFile;
 
 	P_NOTICE(SHORT_PROGRAM_NAME " core running in single-application mode.");
-	P_NOTICE("Serving app     : " << appRoot);
-	P_NOTICE("App type        : " << appType);
-	P_NOTICE("App startup file: " << startupFile);
+	P_NOTICE("Serving app      : " << appRoot);
+	if (!appType.empty()) {
+		P_NOTICE("App type         : " << appType);
+		P_NOTICE("App startup file : " << startupFile);
+		wo->singleAppModeConfig["app_type"] = appType;
+		wo->singleAppModeConfig["startup_file"] = startupFile;
+	} else {
+		P_NOTICE("App start command: " << appStartCommand);
+		wo->singleAppModeConfig["app_start_command"] = appStartCommand;
+	}
 }
 
 static void
@@ -551,39 +592,62 @@ printInfoInThread() {
 }
 
 static void
-dumpDiagnosticsOnCrash(void *userData) {
+dumpOxtBacktracesOnCrash(void *userData) {
+	cerr << oxt::thread::all_backtraces();
+	cerr.flush();
+}
+
+static void
+dumpControllerStatesOnCrash(void *userData) {
 	WorkingObjects *wo = workingObjects;
 	unsigned int i;
 
-	cerr << "### Backtraces\n";
-	cerr << oxt::thread::all_backtraces();
-	cerr.flush();
-
 	for (i = 0; i < wo->threadWorkingObjects.size(); i++) {
 		ThreadWorkingObjects *two = &wo->threadWorkingObjects[i];
-		cerr << "### Request handler state (thread " << (i + 1) << ")\n";
+		cerr << "####### Controller state (thread " << (i + 1) << ") #######\n";
 		cerr << two->controller->inspectStateAsJson();
-		cerr << "\n";
+		cerr << "\n\n";
 		cerr.flush();
 	}
+}
+
+static void
+dumpControllerConfigsOnCrash(void *userData) {
+	WorkingObjects *wo = workingObjects;
+	unsigned int i;
 
 	for (i = 0; i < wo->threadWorkingObjects.size(); i++) {
 		ThreadWorkingObjects *two = &wo->threadWorkingObjects[i];
-		cerr << "### Request handler config (thread " << (i + 1) << ")\n";
+		cerr << "####### Controller config (thread " << (i + 1) << ") #######\n";
 		cerr << two->controller->inspectConfig();
-		cerr << "\n";
+		cerr << "\n\n";
 		cerr.flush();
 	}
+}
 
-	cerr << "### Pool state (simple)\n";
+static void
+dumpPoolStateOnCrash(void *userData) {
+	WorkingObjects *wo = workingObjects;
+
+	cerr << "####### Pool state (simple) #######\n";
 	// Do not lock, the crash may occur within the pool.
 	Pool::InspectOptions options(Pool::InspectOptions::makeAuthorized());
 	options.verbose = true;
 	cerr << wo->appPool->inspect(options, false);
-	cerr << "\n";
+	cerr << "\n\n";
 	cerr.flush();
 
-	cerr << "### mbuf stats\n\n";
+	cerr << "####### Pool state (XML) #######\n";
+	Pool::ToXmlOptions options2(Pool::ToXmlOptions::makeAuthorized());
+	options2.secrets = true;
+	cerr << wo->appPool->toXml(options2, false);
+	cerr << "\n\n";
+	cerr.flush();
+}
+
+static void
+dumpMbufStatsOnCrash(void *userData) {
+	WorkingObjects *wo = workingObjects;
 	cerr << "nfree_mbuf_blockq  : " <<
 		wo->threadWorkingObjects[0].serverKitContext->mbuf_pool.nfree_mbuf_blockq << "\n";
 	cerr << "nactive_mbuf_blockq: " <<
@@ -592,13 +656,6 @@ dumpDiagnosticsOnCrash(void *userData) {
 		wo->threadWorkingObjects[0].serverKitContext->mbuf_pool.mbuf_block_chunk_size << "\n";
 	cerr << "\n";
 	cerr.flush();
-
-	cerr << "### Pool state (XML)\n";
-	Pool::ToXmlOptions options2(Pool::ToXmlOptions::makeAuthorized());
-	options2.secrets = true;
-	cerr << wo->appPool->toXml(options2, false);
-	cerr << "\n\n";
-	cerr.flush();
 }
 
 static void
@@ -653,6 +710,7 @@ initializeNonPrivilegedWorkingObjects() {
 	wo->spawningKitContext = boost::make_shared<SpawningKit::Context>(
 		wo->spawningKitContextSchema);
 	wo->spawningKitContext->resourceLocator = &wo->resourceLocator;
+	wo->spawningKitContext->wrapperRegistry = coreWrapperRegistry;
 	wo->spawningKitContext->randomGenerator = wo->randomGenerator;
 	wo->spawningKitContext->integrationMode = coreConfig->get("integration_mode").asString();
 	wo->spawningKitContext->instanceDir = coreConfig->get("instance_dir").asString();
@@ -713,6 +771,7 @@ initializeNonPrivilegedWorkingObjects() {
 			&wo->singleAppModeConfig,
 			coreSchema->controllerSingleAppMode.translator);
 		two.controller->resourceLocator = &wo->resourceLocator;
+		two.controller->wrapperRegistry = coreWrapperRegistry;
 		two.controller->appPool = wo->appPool;
 		two.controller->shutdownFinishCallback = controllerShutdownFinished;
 		two.controller->initialize();
@@ -821,6 +880,26 @@ initializeSecurityUpdateChecker() {
 	checker->start();
 }
 
+static void
+initializeTelemetryCollector() {
+	TRACE_POINT();
+	WorkingObjects &wo = *workingObjects;
+
+	Json::Value config = coreConfig->inspectEffectiveValues();
+	TelemetryCollector *collector = new TelemetryCollector(
+		coreSchema->telemetryCollector.schema,
+		coreConfig->inspectEffectiveValues(),
+		coreSchema->telemetryCollector.translator);
+	wo.telemetryCollector = collector;
+	for (unsigned int i = 0; i < wo.threadWorkingObjects.size(); i++) {
+		ThreadWorkingObjects *two = &wo.threadWorkingObjects[i];
+		collector->controllers.push_back(two->controller);
+	}
+	collector->initialize();
+	collector->start();
+	wo.shutdownCounter.fetch_add(1, boost::memory_order_relaxed);
+}
+
 static void
 runAdminPanelConnector(AdminPanelConnector *connector) {
 	connector->run();
@@ -968,6 +1047,50 @@ reportInitializationInfo() {
 	}
 }
 
+static void
+initializeAbortHandlerCustomerDiagnostics() {
+	if (!Agent::Fundamentals::abortHandlerInstalled()) {
+		return;
+	}
+
+	Agent::Fundamentals::AbortHandlerConfig::DiagnosticsDumper *diagnosticsDumpers
+		= &Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumpers[0];
+
+	diagnosticsDumpers[0].name = "OXT backtraces";
+	diagnosticsDumpers[0].logFileName = "backtrace_oxt.log";
+	diagnosticsDumpers[0].func = dumpOxtBacktracesOnCrash;
+
+	diagnosticsDumpers[1].name = "controller states";
+	diagnosticsDumpers[1].logFileName = "controller_states.log";
+	diagnosticsDumpers[1].func = dumpControllerStatesOnCrash;
+
+	diagnosticsDumpers[2].name = "controller configs";
+	diagnosticsDumpers[2].logFileName = "controller_configs.log";
+	diagnosticsDumpers[2].func = dumpControllerConfigsOnCrash;
+
+	diagnosticsDumpers[3].name = "pool state";
+	diagnosticsDumpers[3].logFileName = "pool.log";
+	diagnosticsDumpers[3].func = dumpPoolStateOnCrash;
+
+	diagnosticsDumpers[4].name = "mbuf statistics";
+	diagnosticsDumpers[4].logFileName = "mbufs.log";
+	diagnosticsDumpers[4].func = dumpMbufStatsOnCrash;
+
+	Agent::Fundamentals::abortHandlerConfigChanged();
+}
+
+static void
+uninstallAbortHandlerCustomDiagnostics() {
+	if (!Agent::Fundamentals::abortHandlerInstalled()) {
+		return;
+	}
+
+	for (unsigned int i = 0; i < Agent::Fundamentals::AbortHandlerConfig::MAX_DIAGNOSTICS_DUMPERS; i++) {
+		Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumpers[i].func = NULL;
+	}
+	Agent::Fundamentals::abortHandlerConfigChanged();
+}
+
 static void
 mainLoop() {
 	TRACE_POINT();
@@ -978,9 +1101,6 @@ mainLoop() {
 			&& maxCpus <= CPU_SETSIZE;
 	#endif
 
-	Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumper = dumpDiagnosticsOnCrash;
-	Agent::Fundamentals::abortHandlerConfigChanged();
-
 	for (unsigned int i = 0; i < wo->threadWorkingObjects.size(); i++) {
 		ThreadWorkingObjects *two = &wo->threadWorkingObjects[i];
 		two->bgloop->start("Main event loop: thread " + toString(i + 1), 0);
@@ -1064,6 +1184,20 @@ apiServerShutdownFinished(Core::ApiServer::ApiServer *server) {
 	serverShutdownFinished();
 }
 
+static void
+telemetryCollectorAsyncShutdownThreadMain() {
+	WorkingObjects *wo = workingObjects;
+	wo->telemetryCollector->stop();
+	serverShutdownFinished();
+}
+
+static void
+asyncShutdownTelemetryCollector() {
+	oxt::thread(telemetryCollectorAsyncShutdownThreadMain,
+		"Telemetry collector shutdown",
+		512 * 1024);
+}
+
 /* Wait until the watchdog closes the feedback fd (meaning it
  * was killed) or until we receive an exit message.
  */
@@ -1085,8 +1219,7 @@ waitForExitEvent() {
 	TRACE_POINT();
 	if (syscalls::select(largestFd + 1, &fds, NULL, NULL, NULL) == -1) {
 		int e = errno;
-		Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumper = NULL;
-		Agent::Fundamentals::abortHandlerConfigChanged();
+		uninstallAbortHandlerCustomDiagnostics();
 		throw SystemException("select() failed", e);
 	}
 
@@ -1122,6 +1255,9 @@ waitForExitEvent() {
 		if (wo->apiWorkingObjects.apiServer != NULL) {
 			wo->apiWorkingObjects.bgloop->safe->runLater(shutdownApiServer);
 		}
+		if (wo->telemetryCollector != NULL) {
+			asyncShutdownTelemetryCollector();
+		}
 		if (wo->adminPanelConnector != NULL) {
 			wo->adminPanelConnector->asyncShutdown();
 		}
@@ -1133,8 +1269,7 @@ waitForExitEvent() {
 			&fds, NULL, NULL, NULL) == -1)
 		{
 			int e = errno;
-			Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumper = NULL;
-			Agent::Fundamentals::abortHandlerConfigChanged();
+			uninstallAbortHandlerCustomDiagnostics();
 			throw SystemException("select() failed", e);
 		}
 
@@ -1150,8 +1285,7 @@ cleanup() {
 	P_DEBUG("Shutting down " SHORT_PROGRAM_NAME " core...");
 	wo->appPool->destroy();
 
-	Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumper = dumpDiagnosticsOnCrash;
-	Agent::Fundamentals::abortHandlerConfigChanged();
+	uninstallAbortHandlerCustomDiagnostics();
 
 	for (unsigned i = 0; i < wo->threadWorkingObjects.size(); i++) {
 		ThreadWorkingObjects *two = &wo->threadWorkingObjects[i];
@@ -1160,6 +1294,11 @@ cleanup() {
 	if (wo->apiWorkingObjects.apiServer != NULL) {
 		wo->apiWorkingObjects.bgloop->stop();
 	}
+	if (wo->telemetryCollector != NULL
+	&& !coreConfig->get("telemetry_collector_disabled").asBool())
+	{
+		wo->telemetryCollector->runOneCycle(true);
+	}
 	wo->appPool.reset();
 	for (unsigned i = 0; i < wo->threadWorkingObjects.size(); i++) {
 		ThreadWorkingObjects *two = &wo->threadWorkingObjects[i];
@@ -1210,12 +1349,14 @@ runCore() {
 		initializeCurl();
 		initializeNonPrivilegedWorkingObjects();
 		initializeSecurityUpdateChecker();
+		initializeTelemetryCollector();
 		initializeAdminPanelConnector();
 		prestartWebApps();
 
 		UPDATE_TRACE_POINT();
 		warnIfPassengerRootVulnerable();
 		reportInitializationInfo();
+		initializeAbortHandlerCustomerDiagnostics();
 		mainLoop();
 
 		UPDATE_TRACE_POINT();
@@ -1277,7 +1418,9 @@ int
 coreMain(int argc, char *argv[]) {
 	int ret;
 
-	coreSchema = new Schema();
+	coreWrapperRegistry = new WrapperRegistry::Registry();
+	coreWrapperRegistry->finalize();
+	coreSchema = new Schema(coreWrapperRegistry);
 	coreConfig = new ConfigKit::Store(*coreSchema);
 	initializeAgent(argc, &argv, SHORT_PROGRAM_NAME " core",
 		*coreConfig, coreSchema->loggingKit.translator,
@@ -1289,5 +1432,6 @@ coreMain(int argc, char *argv[]) {
 
 	ret = runCore();
 	shutdownAgent(coreSchema, coreConfig);
+	delete coreWrapperRegistry;
 	return ret;
 }
