@@ -60,7 +60,8 @@ static int                first_start = 1;
 ngx_str_t                 pp_schema_string;
 ngx_str_t                 pp_placeholder_upstream_address;
 PP_CachedFileStat        *pp_stat_cache;
-PP_AppTypeDetector       *pp_app_type_detector;
+PsgWrapperRegistry       *psg_wrapper_registry;
+PsgAppTypeDetector       *psg_app_type_detector;
 PsgWatchdogLauncher      *psg_watchdog_launcher = NULL;
 ngx_cycle_t              *pp_current_cycle;
 
@@ -348,7 +349,7 @@ start_watchdog(ngx_cycle_t *cycle) {
     }
 
     if (autogenerated_main_conf->stat_throttle_rate != NGX_CONF_UNSET_UINT) {
-        pp_app_type_detector_set_throttle_rate(pp_app_type_detector,
+        psg_app_type_detector_set_throttle_rate(psg_app_type_detector,
             autogenerated_main_conf->stat_throttle_rate);
     }
 
@@ -367,6 +368,8 @@ start_watchdog(ngx_cycle_t *cycle) {
     psg_json_value_set_ngx_str_ne(w_config, "instance_registry_dir", &autogenerated_main_conf->instance_registry_dir);
     psg_json_value_set_ngx_flag  (w_config, "security_update_checker_disabled", autogenerated_main_conf->disable_security_update_check);
     psg_json_value_set_ngx_str_ne(w_config, "security_update_checker_proxy_url", &autogenerated_main_conf->security_update_check_proxy);
+    psg_json_value_set_ngx_flag  (w_config, "telemetry_collector_disabled", autogenerated_main_conf->disable_anonymous_telemetry);
+    psg_json_value_set_ngx_str_ne(w_config, "telemetry_collector_proxy_url", &autogenerated_main_conf->anonymous_telemetry_proxy);
     psg_json_value_set_ngx_flag  (w_config, "user_switching", autogenerated_main_conf->user_switching);
     psg_json_value_set_ngx_flag  (w_config, "show_version_in_header", autogenerated_main_conf->show_version_in_header);
     psg_json_value_set_ngx_flag  (w_config, "turbocaching", autogenerated_main_conf->turbocaching);
@@ -511,7 +514,10 @@ pre_config_init(ngx_conf_t *cf)
     pp_placeholder_upstream_address.data = (u_char *) "unix:/passenger_core";
     pp_placeholder_upstream_address.len  = sizeof("unix:/passenger_core") - 1;
     pp_stat_cache = pp_cached_file_stat_new(1024);
-    pp_app_type_detector = pp_app_type_detector_new(DEFAULT_STAT_THROTTLE_RATE);
+    psg_wrapper_registry = psg_wrapper_registry_new();
+    psg_wrapper_registry_finalize(psg_wrapper_registry);
+    psg_app_type_detector = psg_app_type_detector_new(psg_wrapper_registry,
+        DEFAULT_STAT_THROTTLE_RATE);
     psg_watchdog_launcher = psg_watchdog_launcher_new(IM_NGINX, &error_message);
 
     if (psg_watchdog_launcher == NULL) {
